/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2003                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita' degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

#include "AmodioLU.hh"
#include "alglin.hh"

#include <iostream>
using namespace std ;

namespace AmodioLUDefine {

  using namespace std ;

  static const valueType ONE   =  1 ;
  static const valueType M_ONE = -1 ;

  /*
  //   _                 _ 
  //  | | ___   __ _  __| |
  //  | |/ _ \ / _` |/ _` |
  //  | | (_) | (_| | (_| |
  //  |_|\___/ \__,_|\__,_|
  */
  void
  AmodioLU::load( indexType         nblock,
                  indexType         n,
                  indexType         q,
                  valueConstPointer AdAu,
                  valueConstPointer H0,
                  valueConstPointer HN,
                  valueConstPointer Hq ) {

    this -> AdAu   = AdAu ;
    this -> H0     = H0 ;
    this -> HN     = HN ;
    this -> Hq     = Hq ;
    this -> nblock = nblock ;
    this -> n      = n ;
    this -> m      = n+q ;

    indexType nnzF   = nblock*n*n ;
    indexType nnzSR  = 2*nnzF ;
    indexType nnzD   = (n+m)*(n+m) ;
    indexType nnzTMP = 2*n*n ;

    indexType nv = nnzF + nnzSR + nnzD + nnzTMP ;
    indexType ni = 2*n*nblock + 2*n+m ;

    baseValue . allocate(nv) ;
    baseIndex . allocate(ni) ;

    SR_blk    = baseValue( nnzSR  ) ;
    D_blk     = baseValue( nnzD   ) ;
    F_blk     = baseValue( nnzF   ) ;
    TMP_blk   = baseValue( nnzTMP ) ;

    perm_blk  = baseIndex( 2*n*nblock ) ;
    ipiv_blk  = baseIndex( n+m ) ;
    ipiv_work = baseIndex( n ) ;

    alglin::copy( nnzSR, AdAu, 1, SR_blk, 1 ) ;

  }

  /*    __ _ _ __ ___ __  ___ __  _   _ 
  //   / _` | '_ ` _ \\ \/ / '_ \| | | |
  //  | (_| | | | | | |>  <| |_) | |_| |
  //   \__,_|_| |_| |_/_/\_\ .__/ \__, |
  //                       |_|    |___/ 
  */
  void
  AmodioLU::amxpy( valueType alpha, valuePointer x, valuePointer y ) const {

    valueType       * yk = y ;
    valueType const * xk = x ;

    for ( indexType k = 0 ; k < nblock ; ++k, xk += n, yk += n ) {
      valueConstPointer Ad = AdAu + 2*k*n*n ;
      valueConstPointer Au = Ad   + n*n ;
      alglin::gemv( "N", n, n, alpha, Ad, n, xk,   1, ONE, yk, 1 ) ;
      alglin::gemv( "N", n, n, alpha, Au, n, xk+n, 1, ONE, yk, 1 ) ;
    }

    alglin::gemv( "No Transpose", m, n,   alpha, H0, m, x,    1, ONE, yk, 1 ) ;
    alglin::gemv( "No Transpose", m, n,   alpha, HN, m, xk,   1, ONE, yk, 1 ) ;
    alglin::gemv( "No Transpose", m, m-n, alpha, Hq, m, xk+n, 1, ONE, yk, 1 ) ;
  }

  /*  
  //    __            _             _         
  //   / _| __ _  ___| |_ ___  _ __(_)_______ 
  //  | |_ / _` |/ __| __/ _ \| '__| |_  / _ \
  //  |  _| (_| | (__| || (_) | |  | |/ /  __/
  //  |_|  \__,_|\___|\__\___/|_|  |_/___\___|
  */
  indexType
  AmodioLU::factorize() {

    /*\
     |  Based on the algorithm 
     |  Pierluigi Amodio and Giuseppe Romanazzi
     |  Algorithm 859: BABDCR: a Fortran 90 package for the Solution of Bordered ABD Linear Systems
     |  ACM Transactions on Mathematical Software, 32, 4, 597—608 (2006)
    \*/
    
    // some constanst
    indexType const nxn = n*n ;
    indexType const nx2 = n*2 ;

    // initialize indices
    valuePointer F    = F_blk ;
    indexPointer perm = perm_blk ;
    
    // 0  1  2  3  4  5  6  7  8  9 10 11 12
    // 0  *  2  *  4  *  6  *  8  * 10  * 12
    // 0  -  *  -  4  -  *  -  8  -  *  - 12
    // 0  -  -  -  *  -  -  -  8  -  -  - 12
    // 0  -  -  -  -  -  -  -  *  -  -  - 12    

    // !!!!!!!!!!!!!!!!!!!!!!!!!
    // !!!! reduction phase !!!!
    // !!!!!!!!!!!!!!!!!!!!!!!!!
    for ( indexType jump = 1 ; jump < nblock ; jump *= 2 ) {

      for ( indexType k = 1 ; k*jump < nblock ; k += 2 ) {
      
        indexType    jC = k*jump ;
        indexType    jL = jC-jump ;
        indexType    jR = min(jC + jump,nblock) ;
        valuePointer S  = SR_blk +  2*jL*nxn ;
        valuePointer RS = SR_blk + (2*jC-1)*nxn ;
        valuePointer R  = SR_blk + (2*jR-1)*nxn ;

        // reshape ( Rb  Sb ) as ( Rb ) and save in ( , ) columns by columns
        //                       ( Sb )
        alglin::copy( 2*nxn, RS, 1, TMP_blk, 1 ) ;
        alglin::gecopy( n, n, TMP_blk,     n, RS,   nx2 ) ;
        alglin::gecopy( n, n, TMP_blk+nxn, n, RS+n, nx2 ) ;

        /*\
         |  reduces a 2x3 block matrix / Sa  Rb     \  to a 1x2 block matrix ( Sa' Rc' )
         |                             \     Sb  Rc /
         |  
         |  by using the following algorithm:
         |
         |    P * / Rb \ = / Rb' \  = / L          \ (U) = / I              \ (L*U) = / I \ (L*U)
         |        \ Sb /   \ Sb' /    \ Sb' U^(-1) /       \ Sb' (L*U)^(-1) /         \ G /
         |
         |  where  and  G = Sb'*U^(-1)*L^(-1).
         |
         |    / -G  I \ * P * / Rb \ = / -G  I \ / I \ (L*U) = /  0  \
         |    \  I  0 /       \ Sb /   \  I  0 / \ G /         \ L*U /
         |
         | ------------------------------------------------------------
         |
         |    / -G  I \ * P * / Sa   0  \ = / -G  I \ / Sa'  Rc'  \ = / Sa''' Rc''' \ 
         |    \  I  0 /       \ 0    Rc /   \  I  0 / \ Sa'' Rc'' /   \ Sa'   Rc'   /
         |
         |  where  Sa''' = Sa'' - G*Sa', Rc''' = Rc'' - G*Rc' and thus
         |
         |    / -G  I \ * P * / Sa  Rb     \ = / Sa'''  0    Rc''' \
         |    \  I  0 /       \     Sb  Rc /   \ Sa'    L*U  Rc'   /
         |  
        \*/
        // factorize RS by means of the LU factorization
        // P * / Rb \ = / L          \ (U)
        //     \ Sb /   \ Sb' U^(-1) /
        indexType INFO = alglin::getrf( nx2, n, RS, nx2, ipiv_work ) ;
        if ( INFO != 0 ) return INFO ;

        // compute G = Sb' U^(-1) L^(-1)
        //  / L          \ (U) = / I                 \ (L*U) =  / I \ (L*U)
        //  \ Sb' U^(-1) /       \ Sb' U^(-1) L^(-1) /          \ G /
        alglin::trsm( "Right", "Lower", "No Transpose", "Unit", n, n, ONE, RS, nx2, RS+n, nx2 ) ;

        // determine the permutation vector
        for ( indexType i = 0 ; i < nx2 ; ++i ) perm[i] = i ;
        for ( indexType i = 0 ; i < n   ; ++i ) {
          indexType ip = ipiv_work[i]-1 ;
          if ( ip != i ) std::swap( perm[i], perm[ip] ) ;
        }

        // P * / Sa   0  \ = / Sa'  Rc'  \ 
        //     \ 0    Rc /   \ Sa'' Rc'' /
        // S = Sa'' and R = Rc''
        valuePointer Sa1 = TMP_blk ;
        valuePointer Rc1 = TMP_blk + nxn ;
        alglin::copy( nxn, S, 1, Sa1, 1 ) ; alglin::zero( nxn, S, 1 ) ;
        alglin::copy( nxn, R, 1, Rc1, 1 ) ; alglin::zero( nxn, R, 1 ) ;
        for ( indexType i = 0 ; i < n ; ++i ) {
          indexType pi = perm[i+n] ;
          if ( pi < n ) alglin::copy( n, Sa1+pi,   n, S+i, n ) ;
          else          alglin::copy( n, Rc1+pi-n, n, R+i, n ) ;
        }

        // / -G  I \ / Sa'  Rc'  \ = / Sa''' Rc''' \ 
        // \  I  0 / \ Sa'' Rc'' /   \ Sa'   Rc'   /
        //
        // where  Sa''' = Sa'' - G*Sa', Rc''' = Rc'' - G*Rc' 
        // where the nonnull rows of Sa' and Rc' are saved in F
        valuePointer Fi = F ;
        valuePointer Gi = RS + n ; // i-th column of G
        for ( indexType i = 0 ; i < n ; ++i, Fi += n, Gi += nx2 ) {
          indexType    pi = perm[i] ;
          valuePointer R_or_S, Ra_or_Sc ;
          if ( pi < n ) { // save nonnull row of Sa' in the column of F
            Ra_or_Sc = Sa1 + pi ;
            R_or_S   = S ;
          } else { // save nonnull row of Rc' in the column of F
            Ra_or_Sc = Rc1 + pi - n ;
            R_or_S   = R ;
          }
          alglin::copy( n, Ra_or_Sc, n, Fi, 1 ) ;
          alglin::ger( n, n, M_ONE, Gi, 1, Fi, 1, R_or_S, n ) ;
        }

        perm += nx2 ;
        F    += nxn ;
      }
    }

    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // !!!! factorization of the last 2 by 2 matrix !!!!
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    valuePointer S = SR_blk ;
    valuePointer R = SR_blk + (2*nblock-1)*nxn ;

    indexType nm = n+m ;
    alglin::gecopy( n, n, S,  n, D_blk,        nm ) ;
    alglin::gecopy( n, n, R,  n, D_blk+n*nm,   nm ) ;

    alglin::gecopy( m, n, H0, m, D_blk+n,      nm ) ;
    alglin::gecopy( m, n, HN, m, D_blk+n+n*nm, nm ) ;

    if ( m > n ) {
      alglin::gezero( n, m-n,        D_blk+nx2*nm,   nm ) ;
      alglin::gecopy( m, m-n, Hq, m, D_blk+nx2*nm+n, nm ) ;
    }

    return alglin::getrf( nm, nm, D_blk, nm, ipiv_blk ) ;
  }
  
  /*             _           
  //   ___  ___ | |_   _____ 
  //  / __|/ _ \| \ \ / / _ \
  //  \__ \ (_) | |\ V /  __/
  //  |___/\___/|_| \_/ \___|
  */                       
  void
  AmodioLU::solve( valuePointer y ) {
    /*\
     |
     |  Purpose
     |  =======
     |
     |  BABDCR_SOLV  solves a babd linear system whose coefficient matrix
     |  has been factorized by BABDCR_FACT. The algorithm consists of three
     |  phases: reduction (by using the subroutine REDUCE_RHS), solution of
     |  the 2 by 2 block system (by using the subroutine DGETRS) and 
     |  back-substitution (by using the subroutine SOLVE_BLOCK).
     |
     |  In input, BABDCR_SOLV requires the coefficient matrix factorized by
     |  the subroutine BABDCR_FACT (see details of this subroutine) and the
     |  right hand side which is stored in the block vector VECT_B in the
     |  following form
     |
     |  VECT_B  = [f(0), f(1), ...., f(NBLOKS)].
     |
     |  The first block element f(0) corresponds to the first row of the
     |  coefficient matrix containing Ba and Bb.
     |
     |  On exit, BABDCR_SOLV gives the solution of the babd linear system.
     |  The solution is stored in the block vector VECT_B.
     |
    \*/
    // some constanst
    indexType const nxn = n*n ;
    indexType const nx2 = 2*n ;

    // initialize indices
    valuePointer F    = F_blk ;
    indexPointer perm = perm_blk ;
    
    // !!!!!!!!!!!!!!!!!!!!!!!!!
    // !!!! reduction phase !!!!
    // !!!!!!!!!!!!!!!!!!!!!!!!!
    indexType jump = 1 ;
    for ( ; jump < nblock ; jump *=2 ) {

      for ( indexType k = 1 ; k*jump < nblock ; k += 2 ) {

        indexType    jC = k * jump ;
        indexType    jL = jC - jump ;
        valuePointer RS = SR_blk + (2*jC-1)*nxn ;
        valuePointer Va = y + jL*n ;
        valuePointer Vb = y + jC*n ;

        /*\
         |   P * / Va \ = / Vp \
         |       \ Vb /   \ Vq /
        \*/
        alglin::copy( n, Va, 1, TMP_blk,   1 ) ; 
        alglin::copy( n, Vb, 1, TMP_blk+n, 1 ) ;

        for ( indexType i = 0 ; i < n ; ++i ) {
          Va[i] = TMP_blk[perm[i+n]] ;
          Vb[i] = TMP_blk[perm[i]] ;
        }
        /*\
         |  / -G  I \ / Vp \ = / Vq-G*Vp \
         |  \  I  0 / \ Vq /   \ Vp      /
        \*/
        alglin::gemv( "No Transpose", n, n, M_ONE, RS + n, nx2, Vb, 1, ONE, Va, 1 ) ;

        perm += nx2 ;
        F    += nxn ;
      }
    }

    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // !!!! 2 by 2 block linear system solution !!!!
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    indexType nm = n+m ;

    // / S  R  0  \ /x(0)\  = b(0)
    // \ H0 HN Hq / \x(N)/  = b(N)
    valuePointer ye = y + (nblock-1) * n ;
    
    // Apply row interchanges to the right hand sides.
    alglin::swap( n, y, 1, ye, 1 ) ;
    alglin::swaps( 1, ye, nm, 0, nm-1, ipiv_blk, 1 ) ;
    alglin::trsv( "Lower", "No transpose", "Unit",     nm, D_blk, nm, ye, 1 ) ;
    alglin::trsv( "Upper", "No transpose", "Non-Unit", nm, D_blk, nm, ye, 1 ) ;
    alglin::swap( n, y, 1, ye, 1 ) ;

    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    // !!!! back-substitution phase !!!!
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    for ( jump /= 2 ; jump > 0 ; jump /= 2 ) {

      indexType k = (nblock-1)/jump ;
      if ( (k & 0x01) == 0 ) --k ; 
    
      for ( ; k > 0 ; k -= 2 ) {
        indexType    jC = k*jump ;
        indexType    jL = jC-jump ;
        indexType    jR = min(jC + jump,nblock) ;
        valuePointer Xm = y + jL*n ;
        valuePointer Xs = y + jR*n ;
        valuePointer V  = y + jC*n ;
        valuePointer RS = SR_blk + (2*jC-1)*nxn ;

        perm -= nx2 ;
        F    -= nxn ;

        /*\
         |  computes the solution Xn (of length n) of the linear system
         |
         |          L*U Xn = Vp - Sp*Xm - Rp*Xs
         |
         |  obtained from the subroutines reduceBlock and reduceRHS applied
         |  to the system (see the subroutine reduceBlock for further details)
         |
         |   ( Sa  Rb     )  ( Xm )   ( Va ).
         |   (     Sb  Rc )  ( Xn ) = ( Vb )
         |                   ( Xs )
         |
         |   (  I    ) * P * ( Sa  Rb     ) ( Xm )   ( Sp  L*U  Rp  ) ( Xm )   ( Vp )
         |   ( -G  I )       (     Sb  Rc ) ( Xn ) = ( Sa'  0   Rc' ) ( Xn ) = ( V' )
         |                                  ( Xs )                    ( Xs )
        \*/
     
        // compute V = Vp - Sp*Xm - Rp*Xs
        for ( indexType i = 0 ; i < n ; ++i )
          V[i] -= alglin::dot( n, F + i*n, 1, perm[i] < n ? Xm : Xs, 1 ) ;

        // solve the system L*U Xn = V
        alglin::trsv( "Lower", "No transpose", "Unit",     n, RS, nx2, V, 1 ) ;
        alglin::trsv( "Upper", "No transpose", "Non Unit", n, RS, nx2, V, 1 ) ;
      }
    }
  }
 
}
