/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2003                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

/**
 * 
 * @mainpage BLOCK LU
 * @date     May 30, 2006
 * @version  1.0
 * @note     first release May 30, 2006
 *
 * @author   Enrico Bertolazzi
 *
 * @par      Affiliation:
 *           Department of Mechanics and Structures Engineering <br>
 *           University of Trento <br>
 *           via Mesiano 77, I -- 38050 Trento, Italy <br>
 *           enrico.bertolazzi@ing.unitn.it
 *
 * @par Preface
 *
 * @par Abstract
 * 
 */

#ifndef AMODIO_LU_HH
#define AMODIO_LU_HH

#include "Malloc.hh"

#ifndef CHAR_TYPE
#define CHAR_TYPE char
#endif

#ifndef REAL_TYPE
#define REAL_TYPE double
#endif

#ifndef INTEGER_TYPE
#define INTEGER_TYPE int
#endif

namespace AmodioLUDefine {

  typedef REAL_TYPE         valueType           ;
  typedef valueType*        valuePointer        ;  
  typedef const valueType*  valueConstPointer   ; 
  typedef valueType&        valueReference      ;
  typedef const valueType&  valueConstReference ;
  
  typedef INTEGER_TYPE      indexType           ;
  typedef indexType*        indexPointer        ;  
  typedef const indexType*  indexConstPointer   ; 
  typedef indexType&        indexReference      ;
  typedef const indexType&  indexConstReference ;

  /*
  //      _                        _ _       _    _   _ 
  //     / \   _ __ ___   ___   __| (_) ___ | |  | | | |
  //    / _ \ | '_ ` _ \ / _ \ / _` | |/ _ \| |  | | | |
  //   / ___ \| | | | | | (_) | (_| | | (_) | |__| |_| |
  //  /_/   \_\_| |_| |_|\___/ \__,_|_|\___/|_____\___/ 
  */
  class AmodioLU {
  private:

    Malloc<valueType> baseValue ;
    Malloc<indexType> baseIndex ;

    AmodioLU(AmodioLU const &) ;
    AmodioLU const & operator = (AmodioLU const &) ;

    indexType nblock ; //!< total number of blocks
    indexType n      ; //!< size of square blocks
    indexType m      ; //!< number final rows (m>=n)
    indexType nnz    ; //!< total number of non zeros

    /*
    //
    //  Matrix structure
    //
    //                  n * nblock
    //    ________________^_________________
    //   /                                  \
    //     n     n     n                        n     q
    //  +-----+-----+-----+----.........-----+-----+-----+    \
    //  |  Ad | Au  |  0  |                  |  0  |  0  | n   |
    //  +-----+-----+-----+             -----+-----+-----+     |
    //  |  0  | Ad  | Au  |  0               |  0  |  0  | n   |
    //  +-----+-----+-----+-----+       -----+-----+-----+     |
    //  |  0  |  0  | Ad  | Au  |            |  0  |  0  | n   |
    //  +-----+-----+-----+-----+       -----+-----+-----+     |   
    //  |                                                :     |
    //  :                                                :      > n * nblock
    //  :                                                :     | 
    //  :                                                :     |
    //  :                                                :     |
    //  :                              +-----+-----+-----+     |
    //  :                              | Au  |  0  |  0  |     |
    //  :                        +-----+-----+-----+-----+     |
    //  :                        |  0  | Ad  | Au  |  0  | n   |
    //  +-----+-----+---......---+-----+-----+=====+=====+    /
    //  |     |     |            |     |     !     |     !
    //  | H0  |  0  |            |     |  0  ! HN  | Hq  ! m
    //  |     |     |            |     |     !     |     !
    //  +-----+-----+---......---+-----+-----+=====+=====+
    //
    */
    valueConstPointer AdAu ;
    valueConstPointer H0   ;
    valueConstPointer HN   ;
    valueConstPointer Hq   ;

    ///////////////////////////////////////////////////////

    valuePointer SR_blk ;
    valuePointer D_blk ;
    valuePointer F_blk ;
    indexPointer ipiv_work ;
    indexPointer ipiv_blk ;
    indexPointer perm_blk ;
    valuePointer TMP_blk ;

  public:

    explicit
    AmodioLU()
    : baseValue("AmodioLU_value")
    , baseIndex("AmodioLU_index")
    { }

    ~AmodioLU() {
      baseValue . free() ;
      baseIndex . free() ;
    }

    void load( indexType         nblk,
               indexType         n,
               indexType         q,
               valueConstPointer AdAu,
               valueConstPointer H0,
               valueConstPointer HN,
               valueConstPointer Hq ) ;

    void amxpy( valueType a, valuePointer x, valuePointer y ) const ;

    indexType factorize() ;

    void solve( valuePointer in_out ) ;

  } ;
}

namespace AmodioLULoad {
  using AmodioLUDefine::AmodioLU ;
}

#endif
