/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2003-2010                                                 |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

/**
 * 
 * @mainpage BLOCK LU
 * @date     February 12, 2010
 * @version  1.0
 * @note     first release May 30, 2006
 *
 * @author   Enrico Bertolazzi
 *
 * @par      Affiliation:
 *           Department of Mechanics and Structures Engineering <br>
 *           University of Trento <br>
 *           via Mesiano 77, I -- 38050 Trento, Italy <br>
 *           enrico.bertolazzi@ing.unitn.it
 *
 * @par Preface
 *
 * @par Abstract
 * 
 */

#include "BlockLU.hh"
#include "Alglin.hh"
#include <limits>

namespace BlockLUDefine {

  using namespace std ;
  
  static const valueType ONE   =  1 ;
  static const valueType M_ONE = -1 ;

  /*   ____  _            _    _    _   _ 
  //  | __ )| | ___   ___| | _| |  | | | |
  //  |  _ \| |/ _ \ / __| |/ / |  | | | |
  //  | |_) | | (_) | (__|   <| |__| |_| |
  //  |____/|_|\___/ \___|_|\_\_____\___/ 
  */
  BlockLU::BlockLU()
  : baseValue("BlockLU_value")
  , baseIndex("BlockLU_index")
  { }

  BlockLU::~BlockLU() {
    baseValue . free() ;
    baseIndex . free() ;
  }

  /*         _ _                 _       
  //    __ _| | | ___   ___ __ _| |_ ___ 
  //   / _` | | |/ _ \ / __/ _` | __/ _ \
  //  | (_| | | | (_) | (_| (_| | ||  __/
  //   \__,_|_|_|\___/ \___\__,_|\__\___|
  */
  void
  BlockLU::allocate( indexType nblock, indexType n, indexType q ) {

    this -> nblock = nblock ;
    this -> n      = n ;
    this -> m      = n+q ;
    this -> N      = nblock*n+m ;
    
    indexType nnzAdH = nblock*(n*(n+m)) ; // blocchi AdH
    indexType nnzAu  = nblock*(n*n)+n*q ; // blocchi Au
    indexType nnzFF  = (nblock-1)*(n*m) ; // blocchi FF
    indexType nnzDD  = m*m ;            ; // blocco  DD

    nnz = nnzAdH + nnzAu + nnzFF + nnzDD ;

    baseValue . allocate( nnz ) ;
    baseIndex . allocate( N   ) ;

    AdH_blk  = baseValue( nnzAdH ) ;
    Au_blk   = baseValue( nnzAu  ) ;
    FF_blk   = baseValue( nnzFF  ) ;
    DD_blk   = baseValue( nnzDD  ) ;
    ipiv_blk = baseIndex( N ) ;
  }

  /*   _                 _ 
  //  | | ___   __ _  __| |
  //  | |/ _ \ / _` |/ _` |
  //  | | (_) | (_| | (_| |
  //  |_|\___/ \__,_|\__,_|
  */
  void
  BlockLU::load( indexType         nblk,
                 indexType         n,
                 indexType         q,
                 valueConstPointer AdAu,
                 valueConstPointer H0,
                 valueConstPointer HN,
                 valueConstPointer Hq ) {

    allocate( nblk, n, q ) ;

    this -> AdAu = AdAu ;
    this -> H0   = H0   ;
    this -> HN   = HN   ;
    this -> Hq   = Hq   ;

    // Initialize structures
    alglin::zero( nnz, AdH_blk, 1 ) ;
    
    // fill matrix
    indexType nm = n+m ;
    for ( indexType k = 0 ; k < nblock ; ++k ) {
      valueConstPointer Ad = AdAu + 2*k*n*n ;
      valueConstPointer Au = Ad   + n*n ;
      alglin::gecopy( n, n, Ad, n, AdH_blk + k*nm*n, nm ) ;
      alglin::gecopy( n, n, Au, n, Au_blk  + k*n*n,  n  ) ;
    }

    alglin::gecopy( m, n, H0, m, AdH_blk + n,  nm ) ;
    alglin::gecopy( m, n, HN, m, DD_blk,       m  ) ;
    alglin::gecopy( m, q, Hq, m, DD_blk + m*n, m  ) ;
  }

  /*    __ _ _ __ ___ __  ___ __  _   _ 
  //   / _` | '_ ` _ \\ \/ / '_ \| | | |
  //  | (_| | | | | | |>  <| |_) | |_| |
  //   \__,_|_| |_| |_/_/\_\ .__/ \__, |
  //                       |_|    |___/ 
  */
  void
  BlockLU::amxpy( valueType alpha, valuePointer x, valuePointer y ) const {

    valueType       * yk = y ;
    valueType const * xk = x ;

    for ( indexType k = 0 ; k < nblock ; ++k, xk += n, yk += n ) {
      valueConstPointer Ad = AdAu + 2*k*n*n ;
      valueConstPointer Au = Ad   + n*n ;
      alglin::gemv( "N", n, n, alpha, Ad, n, xk,   1, ONE, yk, 1 ) ;
      alglin::gemv( "N", n, n, alpha, Au, n, xk+n, 1, ONE, yk, 1 ) ;
    }

    alglin::gemv( "No Transpose", m, n,   alpha, H0, m, x,    1, ONE, yk, 1 ) ;
    alglin::gemv( "No Transpose", m, n,   alpha, HN, m, xk,   1, ONE, yk, 1 ) ;
    alglin::gemv( "No Transpose", m, m-n, alpha, Hq, m, xk+n, 1, ONE, yk, 1 ) ;
  }
  
  /*    __            _             _         
  //   / _| __ _  ___| |_ ___  _ __(_)_______ 
  //  | |_ / _` |/ __| __/ _ \| '__| |_  / _ \
  //  |  _| (_| | (__| || (_) | |  | |/ /  __/
  //  |_|  \__,_|\___|\__\___/|_|  |_/___\___|
  */
  indexType
  BlockLU::factorize() {

    indexType    nm    = n+m ;
    indexType    rowFF = (nblock-1)*n ;
    indexType    INFO ;

    indexPointer ipivk = ipiv_blk ;
    valuePointer AdH   = AdH_blk  ;
    valuePointer Au    = Au_blk   ;
    valuePointer FF    = FF_blk   ;

    for ( indexType k = 0 ; k < nblock-1 ; ++k, ipivk += n, AdH += nm*n, Au += n*n, FF += n ) {

      INFO = alglin::getrf( nm, n, AdH, nm, ipivk ) ; // LU factorization
      if ( INFO != 0 ) return INFO + k*n ;

      valuePointer H  = AdH + n ;
      valuePointer CC = AdH + nm*n + n ;

      for ( indexType i = 0 ; i < n ; ++i ) {
        indexType ip = ipivk[i]-1 ;
        if ( ip != i ) { // detected row exchange
          if ( ip < n ) { // exchange row
            alglin::swap( n, Au + i, n,     Au + ip, n     ) ;
            alglin::swap( m, FF + i, rowFF, FF + ip, rowFF ) ; // last column block
          } else {
            alglin::swap( n, Au + i, n,     CC     + (ip-n), nm ) ;
            alglin::swap( m, FF + i, rowFF, DD_blk + (ip-n), m  ) ; // last column block
          }
        }
      }

      //
      //  +---------+---------+ ........ +--------+
      //  | \    U  |         |          |        |
      //  |    \    |L^(-1)Au |          |L^(-1)FF|
      //  |   L   \ |         |          |        |
      //  |=========|=========|          +========+
      //  |         |         |          |        |
      //  |    H    |   CC*   |          |   DD*  |
      //  |         |         |          |        |
      //  +---------+---------+ ........ +--------+
      //
      //  CC* = CC - H (LU)^(-1) Au
      //  DD* = DD - H (LU)^(-1) FF
      //
      alglin::trsm( "Left",  "Lower", "No Transpose", "Unit", n, n, ONE, AdH, nm, Au, n ) ;
      alglin::trsm( "Left",  "Lower", "No Transpose", "Unit", n, m, ONE, AdH, nm, FF, rowFF ) ;

      alglin::gemm( "No Transpose", "No Transpose", m, n, n, M_ONE, H, nm, Au, n, ONE, CC, nm ) ;
      alglin::gemm( "No Transpose", "No Transpose", m, m, n, M_ONE, H, nm, FF, rowFF, ONE, DD_blk, m ) ;
    }

    // factorize last two block
    INFO = alglin::getrf( nm, n, AdH, nm, ipivk ) ; // LU factorization
    if ( INFO != 0 ) return false ;

    for ( indexType i = 0 ; i < n ; ++i ) {
      indexType ip = ipivk[i]-1 ;
      if ( ip != i ) { // detected row exchange
        if ( ip < n ) { // exchange row
          alglin::swap( m, Au + i, n, Au + ip, n ) ; // last column block
        } else {
          alglin::swap( m, Au + i, n, DD_blk + (ip-n), m ) ; // last column block
        }
      }
    }

    //
    // +---------+---------+
    // | \    U  |         |
    // |    \    |L^(-1)Au |
    // |   L   \ |         |
    // |=========|=========|
    // |         |         |
    // |    H    |   CC*   |
    // |         |         |
    // +---------+---------+
    //
    //  CC* = CC - H (LU)^(-1) Au
    //
    valuePointer H = AdH + n ;
    alglin::trsm( "Left", "Lower", "No Transpose", "Unit", n, m, ONE, AdH, nm, Au, n ) ;
    alglin::gemm( "No Transpose", "No Transpose", m, m, n, M_ONE, H, nm, Au, n, ONE, DD_blk, m ) ;
    
    // factorize last block
    ipivk += n ;
    INFO = alglin::getrf( m, m, DD_blk, m, ipivk ) ; // LU factorization    
    return INFO == 0 ;
  }

  /*             _           
  //   ___  ___ | |_   _____ 
  //  / __|/ _ \| \ \ / / _ \
  //  \__ \ (_) | |\ V /  __/
  //  |___/\___/|_| \_/ \___|
  */
  void
  BlockLU::solve( valuePointer y ) {

    // solve L
    indexType    nm    = n+m ;
    indexType    rowFF = (nblock-1) * n ;
    valuePointer ye    = y + nblock * n ;

    for ( indexType k = 0 ; k < nblock ; ++k ) {
      indexConstPointer ipivk = ipiv_blk + k * n ;
      valueConstPointer AdH   = AdH_blk  + k * (nm*n) ;
      valuePointer      yk    = y        + k * n ;

      // apply permutation
      for ( indexType i = 0 ; i < n ; ++i ) {
        indexType ip = ipivk[i]-1 ;
        if ( ip != i ) { // detected row exchange
          if ( ip < n ) std::swap( yk[i], yk[ip] ) ;
          else          std::swap( yk[i], ye[ip-n] ) ;
        }
      }
      alglin::trsv( "Lower", "No Transpose", "Unit", n, AdH, nm, yk, 1 ) ;
      alglin::gemv( "No Transpose", m, n, M_ONE, AdH + n, nm, yk, 1, ONE, ye, 1 ) ;
    }

    indexConstPointer ipive = ipiv_blk + nblock * n ;
    indexType ok = alglin::getrs( "No Transpose", m, 1, DD_blk, m, ipive, ye, m ) ;

    if ( rowFF > 0 ) alglin::gemv( "No Transpose", rowFF, m, M_ONE, FF_blk, rowFF, ye, 1, ONE, y, 1 ) ;
    if (     m > n ) alglin::gemv( "No Transpose", n, m-n, M_ONE, Au_blk + nblock*n*n, n, ye+n, 1, ONE, ye-n, 1 ) ;

    indexType k = nblock ;
    do {
      --k ;
      valueConstPointer AdH = AdH_blk + k*nm*n ;
      valueConstPointer Au  = Au_blk  + k*n*n  ;
      valuePointer      yk  = y       + k*n    ;

      alglin::gemv( "No Transpose", n, n, M_ONE, Au, n, yk + n, 1, ONE, yk, 1 ) ;
      alglin::trsv( "Upper", "No Transpose", "Non Unit", n, AdH, nm, yk, 1 ) ;      

    } while ( k > 0 ) ;
  }
}
